/*
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright (c) 2011 digitalSTROM.org, Zurich, Switzerland
 *  Author: Roman Khler <roman.koehler@aizo.com>
 *  Complete rewrite by: Andreas Brauchli <andreas.brauchli@aizo.com>
 */

var LOGFILE_NAME = 'timed-events.log',
    LOG = new Logger(LOGFILE_NAME),
    WEEK = ['SU','MO','TU','WE','TH','FR','SA'],
    daytimeEvents = {
        sunrise: 'sunrise',
        sunset: 'sunset'
    };

function pad(c) {
    var i = parseInt(c, 10);
    return (i < 10 ? '0' + i : c.toString());
}

function getDayIdx(d) {
    var i;
    for (i = 0; i < 7; ++i) {
        if (d == WEEK[i]) {
            return i;
        }
    }
    return -1;
}

/**
 * Returns an unused unique identifier (id)
 * @return {String} A unique string
 */
function getId() {
    var ndEvents = Property.getNode('events'),
        events,
        max;

    if (ndEvents) {
        events = ndEvents.getChildren();
        max = 0;
        while (ndEvents.getChild(max.toString())) {
            ++max;
        }
        return max.toString();

    } else {
        return "0";
    }
}

/**
 * Function that splits the daytime from the time string
 * @param time {String} The time string as saved
 * @return {Array|Boolean|String}
 *          Either (0): An array of length 3 containing:
 *              {String} the daytime event,
 *              {Boolean} sign of the delay
 *              {String} delay
 *          Or (1): The time string as input if time is not a daytime
 *          Or (2): false if the time input is not a valid time or daytime
 */
function getDaytime(time) {
    if (isNaN(time[0])) {
        // assume daytime (relative time of day like sunrise, sunset)
        // this is based on the first char of time not being a number

        if (daytimeEvents[time] !== undefined) {
            // allow the keyword without delay
            return [time, true, '00:00:00'];
        }
        var delaySign = true,
            signIdx = -1;
        if ((signIdx = time.indexOf('-')) > 0) {
            delaySign = false; // negative delay
        } else if ((signIdx = time.indexOf('+')) <= 0) {
            LOG.logln('Error: invalid daytime format');
            return false;
        }
        var daytimeEvent = daytimeEvents[time.substr(0, signIdx)];
        if (daytimeEvent === undefined) {
            LOG.logln('Error: invalid daytime event `' + daytimeEvent + '\'');
            return false;
        }
        var delay = time.substr(signIdx + 1);
        return [daytimeEvent, delaySign, delay];
    }
    return time;
}

/**
 * Get todays daytime for a specific daytime event
 * @return {Date} Date object with the given event time set for today's date
 *          or null if no system daytime is set is set or daytime is invalid.
 */
function getDaytimeDate(daytime) {
    var ndGeo = Property.getNode('/config/geodata'),
        res = new Date(),
        evt, t;

    if (daytimeEvents[daytime] === undefined) {
        LOG.logln('Error: invalid daytime `' + daytime + '\'');
        return null;
    }

    if (ndGeo && ndGeo.getChild(daytime)) {
        evt = ndGeo.getChild(daytime).getValue();
        t = evt.split(':');
        res.setHours(parseInt(t[0], 10));
        res.setMinutes(parseInt(t[1], 10));
        res.setSeconds(parseInt(t[2], 10));
        return res;

    } else {
        LOG.logln('Error: device location is not set');
        return null;
    }
}

/**
 * Checks if the system version matches at least a given version
 * @param version Version to check against (in the format X.Y.Z)
 * @return true if the system version is at least the passed version,
     false otherwise, including the case where the version node doesn't exist
 */
function requireSystemVersion(version) {
    var ndSysVersion = Property.getNode('/system/version/version');
    if (ndSysVersion) {
        var sysVersion = ndSysVersion.getValue().split('.');
        var reqVersion = version.split('.');
        var len = sysVersion.length < reqVersion.length ? sysVersion.length : reqVersion.length;
        var i;
        for (i = 0; i < len; ++i) {
            var sys = parseInt(sysVersion[i], 10);
            var req = parseInt(reqVersion[i], 10);
            if (req < sys || (i === len - 1 && req === sys)) {
                return true;
            }
            if (req > sys) {
                return false;
            }
        }
    }
    return false;
}

function requireSystemLocationSupport() {
    if (!requireSystemVersion('1.2.0')) {
        LOG.logln('Warning: this dSS version does not support daytimed events - please update your dSS');
        return false;
    }
    var ndGeo = Property.getNode('/config/geodata');
    if (!ndGeo || !ndGeo.getChild('latitude')) {
        LOG.logln('Warning: device location is not set. Support for daytime based events is disabled');
        return false;
    }
    return true;
}

/**
 * Requeue all daytime based events
 */
function requeueDaytimeEvents() {
    var ndEvents = Property.getNode('events'),
        events,
        i,
        params;

    if (ndEvents) {
        events = ndEvents.getChildren();
        for (i = 0; i < events.length; ++i) {
            params = loadParams(events[i]);
            if (params.isDaytimeEvent && !params.disabled) {
                // requeue won't work since it's queued correctly
                dequeueEvent(events[i]);
                queueEvent(events[i], params);
            }
        }
    }
}

/**
 * Initialization function. Called on dSS startup
 */
function startup() {
    // prepare app
    Property.load();
    LOG.logln('timed-events running');

    var ndVersion = Property.getNode('version');
    if (ndVersion) {
        var version = ndVersion.getValue().split('.');
        if (version.length > 1 && version[0] == '0' && parseInt(version[1], 10) < 9) {
            convertOldEvents();
        } else if (version[0] && version[0] <= 1
                    && (version[1] === undefined
                        || (version[1] <= 0
                            && (version[2] === undefined
                                || (version[2] <= 0
                                    && (version[3] === undefined
                                        || version[3] < 4
        )))))) {
            // convert weekday format on all versions up to 1.0.0.3
            convertWeekdayFormat();
        }
    }

    // requeue events;
    var ndEvents = Property.getNode('events');
    if (ndEvents) {
        var events = ndEvents.getChildren();
        var i;
        for (i = 0; i < events.length; ++i) {
            requeueEvent(events[i]);
        }

    } else {
        createNode('events');
    }

    // subscribe to daytime changes
    Property.setListener('/config/geodata/sunset', requeueDaytimeEvents);

    //set version number
    Property.setProperty('version', '1.0.0.3');
    Property.setFlag('version', 'ARCHIVE', true);
    Property.store();
}

/**
 * Execute actions delayed one by one
 * @param actions {Array} Actions to execute (in the order of the array)
 * @param time {Number} Miliseconds to wait between actions
 * @param callback {Function} (optional) callback to execute after the last action execution
 */
function executeActions(actions, time, callback) {
    var action = actions.shift(),
        target = null,
        callNextAction = function () { executeActions(actions, time, callback); };

    // filter target by presence to avoid dSS crash (bug #1261)
    if (action.group != null) {
        LOG.logln('target zone ' + action.zone + ' group ' + action.group);
        target = getDevices().byZone(action.zone).byGroup(action.group).byPresence(true);

    } else if (action.device != null) {
        LOG.logln('target device ' + action.device);
        target = getDevices().byPresence(true).byDSID(action.device);
    }

    if (target && (!target.length || target.length() > 0)) {
        if (action.scene !== undefined) {
            LOG.logln('Exec: calling scene ' + action.scene);
            target.callScene(parseInt(action.scene, 10));
        }
        else if (action.blink) {
            LOG.logln('Exec: blinking');
            target.blink();
        }
        else if (action.outputValue !== undefined) {
            LOG.logln('Exec: setting ouput ' + action.outputValue);
            target.setValue(parseInt(action.outputValue, 10));
        }
    } else {
        LOG.logln('Warning: empty target (devices probably offline)');
    }

    if (actions.length > 0) {
        // LOG.logln('debug: Scheduling next '+ actions.length +' params in '+ time +'ms');
        setTimeout(callNextAction, time);

    } else if (callback !== undefined) {
        callback();
    }
} // executeActions

/**
 * Parses the recurrence object and generates an array of strings, each
 * containing a valid recurrence rule of FUTURE DATES.
 * NON RECURRING EVENTS IN THE PAST WILL NOT SHOW UP IN THE RETURN ARRAY
 * The dSS only accepts iCal dates if one provides both rrule and start time
 * @param params {Object} The event's parameters object
 * @return {Array} An array of {Object} of type:
 *      iCalStartTTime {String} The starting date/time iCal string like YYYYMMDDTHHMMSS
 *      iCalRRule {String} The recurrence rule iCal string like FREQ=DAILY;COUNT=1
 */
function generateIcalRules(params) {
    var rules = [], // return array
        time = params.time.split(':'), // time array (delay array daytime event)
        timestring = '', // timestring used in the iCal rule
        MS_PER_WEEK = 7 * 24 * 60 * 60 * 1000,
        now = new Date(),
        // the date object with scheduled date/time scheduled one week in the
        // past to circumvent a possible libical bug causing events not to be
        // executed on certain days. Past events are simply skipped when
        // scheduling so this causes no averse effects
        d = new Date(now.getTime() - MS_PER_WEEK),
        // use compatibility mode with old dSS versions (< 1.0.5)
        // The problem with that, is that it doesn't do daylight savings
        use_utc_time = !requireSystemVersion('1.0.5'),

        // loop variables:
        i,
        rrule = null,
        iCalStartTime = null;

    if (params.isDaytimeEvent) {
        var daytime = getDaytime(params.time);
        time = daytime[2].split(':');
        var delay = [
            parseInt(time[0], 10),
            parseInt(time[1], 10),
            parseInt(time[2], 10)
        ];
        var delayMs = ((((delay[0] * 60) + delay[1]) * 60) + delay[2]) * 1000;
        if (!daytime[1]) {
            // invert delay on "before" event
            delayMs = -delayMs;
        }
        d = getDaytimeDate(daytime[0]);
        if (d === null) {
            return [];
        }
        // See declaration of d for MS_PER_WEEK explanation
        d.setTime(d.getTime() + delayMs - MS_PER_WEEK);

    } else {
        d.setHours(parseInt(time[0], 10));
        d.setMinutes(parseInt(time[1], 10));
        d.setSeconds(parseInt(time[2], 10));
    }
    // This is not daytime event

    if (use_utc_time) {
        // run in compatibility mode
        timestring =
            pad(d.getUTCHours())
            + pad(d.getUTCMinutes())
            + pad(d.getUTCSeconds())
            + 'Z';
         LOG.logln('Warning: event scheduled in UTC time (no daylight savings). Please update your dSS');
    } else {
        // use local time
        timestring =
            pad(d.getHours())
            + pad(d.getMinutes())
            + pad(d.getSeconds());
    }

    if (params.recurring) {

        iCalStartTime =
            d.getFullYear()
            + pad(d.getMonth() + 1) // goes from 0..11
            + pad(d.getDate())
            + 'T'
            + timestring;

        var weekdays = [];
        for (i = 0; i < params.recurrence.length; ++i) {
            weekdays.push(WEEK[params.recurrence[i]]);
        }
        rrule = 'FREQ=WEEKLY;BYDAY=' + weekdays;

        rules.push({
            iCalRRule: rrule,
            iCalStartTTime: iCalStartTime
        });

    } else {
        rrule = 'FREQ=DAILY;COUNT=1';
        for (i = 0; i < params.recurrence.length; ++i) {
            var dt = params.recurrence[i].split('-');
            d.setFullYear(parseInt(dt[0], 10));
            d.setMonth(parseInt(dt[1], 10) - 1);
            d.setDate(parseInt(dt[2], 10));
            if (d < now) {
                // skip past event
                LOG.logln('Debug: skipping past event ' + d.toString());
                continue;
            }
            iCalStartTime = params.recurrence[i].replace(/-/g, '') + 'T' + timestring;
            rules.push({
                iCalRRule: rrule,
                iCalStartTTime: iCalStartTime
            });
        }
    }

    return rules;
}

/**
 * Remove all queued occurences of this event
 * @param ndEvent The event node of the event to reqeue
 */
function dequeueEvent(ndEvent) {
    var ndQueue = ndEvent.getChild('queueIDs');
    if (!ndQueue) {
        return;
    }
    var queue = ndQueue.getChildren();
    var i;
    for (i = 0; i < queue.length; ++i) {
        var propNode = Property.getNode(
            '/system/EventInterpreter/ScheduledEvents/' + queue[i].getValue()
        );
        if (propNode) {
            propNode.getParent().removeChild(propNode);
        }
        ndQueue.removeChild(queue[i]);
    }
}

/**
 * Queue an event
 * @param ndEvent The event node of the event to reqeue
 * @param params (optional) pre-loaded params of the node
 */
function queueEvent(ndEvent, params) {
    var qIdPath = 'events/' + ndEvent.name + '/queueIDs';
    var ndQueue = ndEvent.getChild('queueIDs') || createNode(qIdPath); // just here to ensure the node is created
    var i;

    // raise the event and store the queue id
    if (!params) {
        params = loadParams(ndEvent);
    }
    if (params.disabled) {
        return;
    }
    LOG.logln('Debug: queueing events for event ' + params.idname);

    var ical = generateIcalRules(params);
    for (i = 0; i < ical.length; ++i) {
        var timerEvent = new TimedICalEvent(
            'timed-events',
            ical[i].iCalStartTTime,
            ical[i].iCalRRule,
            {
                event_id: params.id,
                action_type: 'execute'
            }
        );
        LOG.logln('Debug: - ' + ical[i].iCalStartTTime);
        Property.setProperty(qIdPath +'/'+ i, timerEvent.raise());
    }
}

/**
 * Check if all occurences of an event are in the queue, else requeue them
 * @param ndEvent The event node of the event to reqeue
 * @param params (optional) pre-loaded params of the node
 */
function requeueEvent(ndEvent, params) {
    var ndQueue = ndEvent.getChild('queueIDs'),
        i, queue, propNode;

    if (!ndQueue) {
        queueEvent(ndEvent, params);

    } else {
        queue = ndQueue.getChildren();
        if (queue.length == 0) {
            queueEvent(ndEvent, params);

        } else {
            for (i = 0; i < queue.length; ++i) {
                propNode = Property.getNode(
                    '/system/EventInterpreter/ScheduledEvents/' + queue[i].getValue()
                );
                if (!propNode) {
                    // if any node doesn't exists, delete and requeue them all
                    dequeueEvent(ndEvent);
                    queueEvent(ndEvent, params);
                    break;
                }
            }
        }
    }
}

/**
 * Fetches the parameters of this event from the property tree.
 * @param eventNode {Object} Property tree node of the event
 * @return {Object} Object containing the read out values of the event
 */
function loadParams(eventNode) {
    LOG.logln('Loading params for node with id: ' + eventNode.name);
    var name = eventNode.getChild('name');
    if (!name) {
        LOG.logln('Error: no event name');
        return null;
    }

    var params = {
        id: eventNode.name,
        idname: eventNode.name + ' (' + name.getValue() + ')',
        name: name.getValue(),
        time: eventNode.getChild('time').getValue(),
        actions: [],
        recurrence: []
    };
    var ndDisabled = eventNode.getChild('disabled');
    params.disabled = (ndDisabled ? !! ndDisabled.getValue() : false);

    // set daytime flag if the time is sunrise+hh:mm:ss or sunset-hh:mm:ss
    // check true if the first char of time is not a number
    params.isDaytimeEvent = (isNaN(params.time[0]));
    if (params.isDaytimeEvent && !requireSystemLocationSupport()) {
        params.disabled = true;
    }

    // LOG.logln('Debug: Loading actions');
    var ndActions = eventNode.getChild('actions');
    var loadValue = function(action, ndAction, name) {
        var c = ndAction.getChild(name);
        if (c) {
            action[name] = c.getValue();
            return true;
        }
        return false;
    };
    if (ndActions) {
        var actions = ndActions.getChildren();
        var a;
        for (a = 0; a < actions.length; ++a) {
            var curAction = {};
            if (loadValue(curAction, actions[a], 'device')) {}
            else if (loadValue(curAction, actions[a], 'group') && loadValue(curAction, actions[a], 'zone')) {}
            else {
                LOG.logln('Error: event '+ params.idname +' has an action with no target');
                continue;
            }
            if (loadValue(curAction, actions[a], 'blink')) {}
            else if (loadValue(curAction, actions[a], 'scene')) {}
            else if (loadValue(curAction, actions[a], 'outputValue')) {}
            else {
                LOG.logln('Error: event '+ params.idname +' has an action with invalid action to take');
                continue;
            }
            params.actions.push(curAction);
            // LOG.logln('Debug: action: '+ JSON.stringify(curAction));
        }
    } else {
        LOG.logln('Error: event '+ params.idname +' has no actions set');
    }

    // LOG.logln('Debug: loading recurrences');
    var ndRecurring = eventNode.getChild('recurring');
    params.recurring = (ndRecurring ? !!ndRecurring.getValue() : false);

    if (params.recurring) {
        // weekdays
        var rec = eventNode.getChild('weekdays').split(','),
            i = 0;

        params.recurrence = [];

        for (i = 0; i < rec.length; ++i) {
            try {
                params.recurrence.push(parseInt(rec[i], 10));
            } catch(e){
                LOG.logln('Error: event '+ params.idname +' has an invalid weekday index: ' + rec[i]);
            }
        }

    } else {
        var ndRecurrence = eventNode.getChild('recurrence');
        if (ndRecurrence) {
            var recurrences = ndRecurrence.getChildren()
                w = 0,
                d = 0;

            // dates
            for (d = 0; d < recurrences.length; ++d) {
                params.recurrence.push(recurrences[d].getValue());
            }
            // LOG.logln('Debug: recurrences: '+ JSON.stringify(params.recurrence));
        } else {
            LOG.logln('Error: event '+ params.idname +' has no recurrence set');
        }
    }
    // LOG.logln('Debug: params loaded');
    return params;
}

/**
 * Updates a node if it exists or creates a new archived property node otherwise
 * @param path {String} Relative path
 * @param value {String|Boolean|Integer} Value to set
 */
function updateOrCreateNode(path, value) {
    var nd = Property.getNode(path);
    if (nd) {
        nd.setValue(value);
    } else {
        createNode(path, value);
    }
}

/**
 * Creates a new archived property node with given value
 * @param path {String} Relative path
 * @param value {String|Boolean|Integer} Value to set
 * @return {Object} The newly created node
 */
function createNode(path, value) {
    var nd = new Property(path);
    if (value !== undefined) {
        nd.setValue(value);
    }
    nd.setFlag('ARCHIVE', true);
    return nd;
}

/**
 * Stores _VALID_ parameters into the property tree.
 * Use checkParams(params) to validate them first!
 * Make sure to call Property.store() after calling this method
 * @param params {Object} Object to store
 * @return {Object} The event node
 */
function storeParams(params) {
    var eventPath = 'events/' + params.id;
    var actionPath = eventPath + '/actions';
    var i, c;

    var ndEvent = Property.getNode(eventPath) || createNode(eventPath);

    updateOrCreateNode(eventPath + '/id', params.id);
    updateOrCreateNode(eventPath + '/name', params.name);
    updateOrCreateNode(eventPath + '/time', params.time);
    updateOrCreateNode(eventPath + '/disabled', !!params.disabled);

    // actions
    var ndActions = ndEvent.getChild('actions');
    if (ndActions) {
        // clean out actions, as indices might not match up
        ndEvent.removeChild(ndActions);
    }
    ndActions = createNode(actionPath);
    // populate actions
    for (i = 0; i < params.actions.length; ++i) {
        var p = actionPath + '/' + i;
        createNode(p);

        if (params.actions[i].zone !== undefined) {
            createNode(p + '/zone', params.actions[i].zone);
        }
        if (params.actions[i].group !== undefined) {
            createNode(p + '/group', params.actions[i].group);
        }
        if (params.actions[i].device) {
            createNode(p + '/device', params.actions[i].device);
        }
        if (params.actions[i].blink) {
            createNode(p + '/blink', true);
        }
        if (params.actions[i].scene !== undefined) {
            createNode(p + '/scene', params.actions[i].scene);
        }
        if (params.actions[i].outputValue !== undefined) {
            createNode(p + '/outputValue', params.actions[i].outputValue);
        }
    }

    // recurrence
    updateOrCreateNode(eventPath + '/recurring', params.recurring);

    var ndRecurrence = ndEvent.getChild('recurrence');
    if (params.recurring) {
        if (ndRecurrence) {
            // remove date recurrence node
            ndEvent.removeChild(ndRecurrence);
        }
        updateOrCreateNode(eventPath + '/weekdays', params.recurrence.join(','));

    } else {
        var ndWeekdays = ndEvent.getChild('weekdays');
        if (ndWeekdays) {
            // remove weekdays
            ndEvent.removeChild(ndWeekdays);
        }

        if (ndRecurrence) {
            // weed out current dates
            ndEvent.removeChild(ndRecurrence);
        }
        createNode(eventPath + '/recurrence');

        // populate date nodes
        for (i = 0; i < params.recurrence.length; ++i) {
            createNode(eventPath + '/recurrence/' + i, params.recurrence[i]);
        }
    }
    return ndEvent;
} // storeParams

/**
 * Checks if parameters conform to what is expected
 * @param eventParams {Object} Object to check
 * @return {Object} Conform object or null on error
 */
function checkParams(eventParams) {
    if (!eventParams.name) {
        LOG.logln('Error: event has no name');
        return null;
    }

    if (!eventParams.actions || eventParams.actions.length === 0) {
        LOG.logln('Error: event has no actions');
        return null;
    }

    if (!eventParams.recurrence || eventParams.recurrence.length === 0) {
        LOG.logln('Error: event has no recurrence');
        return null;
    }

    if (!eventParams.time) {
        LOG.logln('Error: event has no time set');
        return null;
    }

    var actions = [];
    var i;
    for (i = 0; i < eventParams.actions.length; ++i) {
        var a = eventParams.actions[i];
        if (!a.device && isNaN(a.group)) { // isNan(undefined) === true
            LOG.logln('Error: no or invalid group or device defined as target on action ' + i);
            return null;
        }
        if (a.blink === undefined && a.scene === undefined && a.outputValue === undefined) {
            LOG.logln('Error: no action defined at index ' + i);
            return null;
        }
        if (
            (a.scene !== undefined && isNaN(a.scene))
            || (a.group !== undefined && isNaN(a.group))
            || (a.outputValue !== undefined && (
                isNaN(a.outputValue) || a.outputValue < 0 || a.outputValue > 255
            ))
        ) {
            LOG.logln('Error: invalid action defined at index ' + i);
            return null;
        }

        var action = {};
        if (a.zone !== undefined) {
            action.zone = parseInt(a.zone, 10);
        }
        if (a.group !== undefined) {
            action.group = parseInt(a.group, 10);
        }
        if (a.device !== undefined) {
            action.device = a.device;
        }
        if (a.blink) {
            action.blink = true;
        }
        if (a.scene !== undefined) {
            action.scene = parseInt(a.scene, 10);
        }
        if (a.outputValue !== undefined) {
            action.outputValue = parseInt(a.outputValue, 10);
        }
        actions.push(action);
    }

    var recurring = !!eventParams.recurring;
    var recurrence = [];
    var d;
    if (recurring) {
        for (i = 0; i < eventParams.recurrence.length; ++i) {
            d = parseInt(eventParams.recurrence[i], 10);
            if (isNaN(d) || d < 0 || d > 6) {
                LOG.logln('Warning: wrong weekday index `'+ d +'\', skipping');
                continue;
            }
            recurrence.push(d);
        }
    } else {
        for (i = 0; i < eventParams.recurrence.length; ++i) {
            d = eventParams.recurrence[i];
            var date = d.split('-');
            if (date.length !== 3 || isNaN(date[0]) || isNaN(date[1]) || isNaN(date[2])
                || date[0].length < 4 || date[1] == 0 || date[1] > 12
                || date[2] == 0 || date[2] > 31
            ) {
                LOG.logln('Warning: invalid date `'+ d +'\'');
                continue;
            }
            recurrence.push(d);
        }
    }
    if (recurrence.length == 0) {
        LOG.logln('Error: no valid recurrence');
        return null;
    }
    var time = eventParams.time;
    var daytime = getDaytime(eventParams.time);
    var daytimeEvent = '';
    var isDaytimeEvent = false;
    if (daytime === false) {
        return null; // error was logged in getDaytime
    } else if (typeof(daytime) !== 'string') {
        isDaytimeEvent = true;
        daytimeEvent = daytime[0] + (daytime[1] ? '+' : '-');
        time = daytime[2];
        if (!requireSystemLocationSupport()) {
            eventParams.disabled = true;
        }
    }
    var t = time.split(':');
    var inv = (t.length !== 3);
    for (i = 0; !inv && i < 3; ++i) {
        if (!isNaN(t[i])) {
            t[i] = parseInt(t[i], 10);
            inv = inv || t[i] < 0 || t[i] > 59;
        } else {
            inv = true;
        }
    }
    inv = inv || (!isDaytimeEvent && t[0] > 23);
    if (inv) {
        LOG.logln('Error: invalid time format');
        return null;
    }
    time = daytimeEvent + pad(t[0]) + ':' + pad(t[1]) + ':' + pad(t[2]);

    // create the valid params object with all fields except id and idname
    var params = {
        name: eventParams.name,
        actions: actions,
        disabled: !!eventParams.disabled,
        recurring: recurring,
        recurrence: recurrence,
        isDaytimeEvent: isDaytimeEvent,
        time: time
    };
    // add id, idname fields if they were supplied
    if (eventParams.id) {
        params.id = eventParams.id;
        params.idname = params.id + ' (' + params.name + ')';
    }

    return params;
} // checkParams

/**
 * Add a new event
 * @param params {Object} The new event is created from these parameters
 */
function addEvent(eventParams) {
    // check eventParams and set params for serialization
    var params = checkParams(eventParams),
        dequeue = false,
        // tmpId is used to communicate the id of the newly created event
        // to the client. Since there is no way to specify a return value
        // for an event, the client will be able to specify a temporary id
        // tmpId and poll for that id. When it's created, it'll contain the
        // definitive id of the created event.
        // The client is charged with deleting this node
        tmpId;

    if (!params) {
        return;
    }

    if (params.id === undefined || params.id.substr(0, 3) === 'tmp') {
        if (params.id !== undefined) {
            tmpId = params.id;
        }
        params.id = getId();
        params.idname = params.id + ' (' + params.name + ')';
    } else {
        // event exists, dequeue it first
        dequeue = true;
    }

    var ndEvent = storeParams(params);

    if (dequeue) {
        dequeueEvent(ndEvent);
    }
    if (!params.disabled) {
        queueEvent(ndEvent, params);
    }

    if (tmpId !== undefined) {
        var evt = new Event('timed-events.create.' + tmpId, { id: params.id });
        evt.raise();
    }

    Property.store();
}

function executeEvent(eventNode, params) {
    if (!params.disabled) {
        LOG.logln(
            'Processing event '+ params.idname
            + ' (' + params.actions.length +' actions)'
        );
        executeActions(params.actions, 500, function() {
            disableOnLastOccurence(eventNode, params);
        });
    }
}

function disableOnLastOccurence(eventNode, params) {
    // Disable irregular event if it's its last occurrence
    if (!params.recurring) {
        if (eventNode.getChild('queueIDs').getChildren().length === 1) {
            enableEvent(eventNode, params, false);

            // notify UI about changes
            var updateEvent = new Event('timed-events.disable', { id: params.id });
            updateEvent.raise();
        }
    }
}

function deleteEvent(eventNode, params) {
    LOG.logln('Delete: event '+ params.idname);

    dequeueEvent(eventNode);
    eventNode.getParent().removeChild(eventNode);
    Property.store();
}

/**
 * Enable or disable a timed event
 * @param enable {Boolean} True to enable, false to disable
 */
function enableEvent(eventNode, params, enable) {
    LOG.logln((enable ? 'Enabling' : 'Disabling') + ' event ' + params.idname);

    params.disabled = !enable;
    updateOrCreateNode('events/' + params.id + '/disabled', !enable);
    if (enable) {
        requeueEvent(eventNode, params);
    } else {
        dequeueEvent(eventNode);
    }
    Property.store();
}

function main() {
    if (raisedEvent.name == 'running') {
        startup();
        return;
    }
    /*
    LOG.logln('Debug: timed-events main call with action type: '
                + raisedEvent.parameter.action_type
                + (raisedEvent.parameter.params === undefined ? ''
                    : ' and params: '
                    + raisedEvent.parameter.params.toString()
                )
    );
    */
    var params;
    if (raisedEvent.parameter.action_type == 'add') {
        // Add a new event
        if (raisedEvent.parameter.params === undefined) {
            LOG.logln('Error: missing event parameters');
            return;
        }
        LOG.logln('Adding event with parmeters: '+ raisedEvent.parameter.params);
        try {
            params = JSON.parse(raisedEvent.parameter.params);
            addEvent(params);
        } catch(e) {
            LOG.logln('Error: Failed to parse params');
        }

    } else {
        // Operations that involve a specific event's parameters
        var ndEvent = Property.getNode('events/' + raisedEvent.parameter.event_id);
        if (ndEvent) {
            params = loadParams(ndEvent);

            if (raisedEvent.parameter.action_type == 'enable') {
                enableEvent(ndEvent, params, true);
            }
            else if (raisedEvent.parameter.action_type == 'disable') {
                enableEvent(ndEvent, params, false);
            }
            else if (raisedEvent.parameter.action_type == 'delete') {
                deleteEvent(ndEvent, params);
            }
            else if (raisedEvent.parameter.action_type == 'execute') {
                executeEvent(ndEvent, params);
            }
        } else {
            LOG.logln('Error: node with id '+ raisedEvent.parameter.event_id +' doesn\'t exist');
        }
    }
} // main

main();

